"""Physics engine interfaces for sensors."""
from typing import Tuple, List
from abc import abstractmethod

import numpy as np
from scipy.spatial.transform import Rotation
import pybullet

class LinkIdNotFoundError(Exception):
    pass


class PhysicsEngineInterface():
    """Physics engine interface for sensors.

    This abstract class defines interfaces that a physics engine must define.
    """

    def __init__(self):
        pass

    @abstractmethod
    def get_obstacle_pose(self, *args, **kwargs) -> Tuple[List[float], List[float]]:
        pass

    @abstractmethod
    def get_obstacle_velocity(self, *args, **kwargs) -> Tuple[List[float], List[float]]:
        pass

    @abstractmethod
    def get_goal_pose(self, goal_id: int) -> Tuple[List[float]]:
        pass

    @abstractmethod
    def get_goal_velocity(self, goal_id: int) -> Tuple[List[float]]:
        pass

    @abstractmethod
    def get_link_position(self, *args) -> np.ndarray:
        pass

    @abstractmethod
    def get_link_orientation(self, *args) -> np.ndarray:
        pass

    def ray_cast(
        self,
        ray_start: np.ndarray,
        ray_end: np.ndarray,
        ray_index: int,
        ray_length: float
    ) -> np.ndarray:
        raise NotImplementedError(
            f"Clearing of visualization of lines not implemented for {type(self)}."
        )

    def clear_visualizations(self) -> None:
        raise NotImplementedError(
            f"Clearing of visualization of lines not implemented for {type(self)}."
        )

    def add_visualization_line(self, *args) -> None:
        raise NotImplementedError(
            f"Visualization of lines not implemented for {type(self)}."
        )


class PybulletInterface(PhysicsEngineInterface):
    """
    Physics engine interface for bullet.
    """
    def extract_link_id(self, robot, link_name: str) -> int:
        number_links = pybullet.getNumJoints(robot)
        joint_names = []
        for i in range(number_links):
            joint_name = pybullet.getJointInfo(robot, i)[1].decode("UTF-8")
            joint_names.append(joint_name)
            if joint_name == link_name:
                return i
        raise LinkIdNotFoundError(
            f"Link with name {link_name} not found. "
            f"Possible links are {joint_names}"
        )


    def get_obstacle_pose(self, obst_id: int, obst_name: str, movable: bool = False) -> None:
        position, orientation = pybullet.getBasePositionAndOrientation(obst_id)
        return position, orientation

    def get_obstacle_velocity(self, obst_id: int, obst_name: str, movable: bool = False) -> None:
        linear, angular = pybullet.getBaseVelocity(obst_id)
        return linear, angular

    def get_goal_pose(self, goal_id: int) -> Tuple[List[float]]:
        position, orientation = pybullet.getBasePositionAndOrientation(goal_id)
        return position, orientation

    def get_goal_velocity(self, goal_id: int) -> Tuple[List[float]]:
        linear, angular = pybullet.getBaseVelocity(goal_id)
        return linear, angular

    def get_link_position(self, robot, link_id) -> np.ndarray:
        link_position = np.array(pybullet.getLinkState(robot, link_id)[0])
        return link_position

    def get_link_orientation(self, robot, link_id) -> np.ndarray:
        link_orientation = np.array(pybullet.getLinkState(robot, link_id)[1])
        return link_orientation

    def ray_cast(
        self,
        ray_start: np.ndarray,
        ray_end: np.ndarray,
        ray_index: int,
        ray_length: float
    ) -> np.ndarray:
        lidar = pybullet.rayTest(ray_start, ray_end)
        return lidar[0][2] * ray_length

    def clear_visualizations(self) -> None:
        pybullet.removeAllUserDebugItems()

    def add_visualization_line(
        self, start_point: Tuple[float], end_point: Tuple[float]
    ) -> None:
        pybullet.addUserDebugLine(start_point, end_point)

class MujocoInterface(PhysicsEngineInterface):
    """
    Physics engine interface for mujoco.
    """
    def extract_link_id(self, robot, link_name: str) -> int:
        return self._data.body(link_name).id

    def set_data(self, data) -> None:
        """Set pointer to mujoco data to sensor."""
        self._data = data

    def get_link_position(self, robot, link_id) -> np.ndarray:
        link_position = self._data.xpos[link_id]
        return link_position

    def get_link_orientation(self, robot, link_id) -> np.ndarray:
        link_orientation_matrix = np.reshape(self._data.xmat[link_id], (3, 3))
        return Rotation.from_matrix(link_orientation_matrix).as_quat()

    def get_obstacle_pose(self, obst_id: int, obst_name: str, movable: bool = False) -> Tuple[List[float], List[float]]:
        if movable:

            free_joint_data = self._data.jnt(f"freejoint_{obst_name}").qpos
            print(free_joint_data)
            return free_joint_data[0:3].tolist(), free_joint_data[3:].tolist()
        pos = self._data.body(obst_name).xpos
        ori = self._data.body(obst_name).xquat
        return pos.tolist(), ori.tolist()

    def get_goal_pose(self, goal_id: int) -> Tuple[List[float]]:
        pos = self._data.site(goal_id).xpos
        goal_rotation = np.reshape(self._data.site(goal_id).xmat, (3, 3))
        ori = Rotation.from_matrix(goal_rotation).as_quat()
        return pos.tolist(), ori.tolist()

    def get_goal_velocity(self, goal_id: int) -> Tuple[List[float]]:
        return [0, 0, 0], [0, 0, 0]

    def ray_cast(
        self,
        ray_start: np.ndarray,
        ray_end: np.ndarray,
        ray_index: int,
        ray_length: float
    ) -> np.ndarray:
        if self._data.sensordata[ray_index] < 0:
            ray_value = ray_length - (0.01 / 2)
        else:
            ray_value = self._data.sensordata[ray_index] - (0.01 / 2)
        return ray_value

    def get_obstacle_velocity(self, obst_id: int, obst_name: str, movable: bool = False) -> None:
        raise NotImplementedError(
            "Obstacle velocity not implemented for mujoco."
        )



import pybullet as p
from typing import List
import gymnasium as gym
import numpy as np
import logging

from urdfenvs.urdf_common.generic_robot import GenericRobot
from urdfenvs.urdf_common.generic_robot import ControlMode


class BicycleModel(GenericRobot):
    """Bicycle model for car like vehicles.

    The bicycle model in velocity mode takes as inputs the forward velocity
    and the steering position. The latter seems counter-intuitive but is more
    common for bicycle models.


    Attributes
    ----------

    _wheel_radius : float
        The radius of the actuated wheels.
    _spawn_offset : np.ndarray
        The offset by which the initial position must be shifted to align
        observation with that position.
    _scaling: float
        The size scaling in which the urdf should be spawned.
        This also effects the dynamics of the system.
    """
    _scaling: float
    _wheel_radius: float
    _wheel_distance: float
    _spawn_offset: np.ndarray
    _facing_direction: str

    def __init__(
            self,
            urdf: str,
            mode: ControlMode,
            actuated_wheels: List[str],
            steering_links: List[str],
            wheel_radius: float,
            wheel_distance: float,
            spawn_offset: np.ndarray = np.array([-0.435, 0.0, 0.05]),
            facing_direction: str = 'x',
            scaling: float = 1.0

        ):
        self._scaling = scaling
        self._wheel_radius = wheel_radius
        self._facing_direction = facing_direction
        self._wheel_distance = wheel_distance
        self._spawn_offset = spawn_offset
        self._steering_links = steering_links
        self._actuated_wheels = actuated_wheels
        self._wheel_radius = wheel_radius
        self._wheel_joints = []
        self._steering_joints = []
        super().__init__(2, urdf, mode)

    def ns(self) -> int:
        """Returns the number of degrees of freedom.

        This is needed as number of actuated joints `_n` is lower that the
        number of degrees of freedom for bycycle models.
        """
        return self.n() + 1

    def extract_joint_ids(self) -> None:
        """Automated extraction of joint ids

        Extract joint ids by the joint names.

        """
        if not hasattr(self, "_steering_joints"):
            return
        if hasattr(self, "_robot"):
            self._robot_joints = []
            self._castor_joints = []
            num_joints = p.getNumJoints(self._robot)
            for name in self._joint_names:
                for i in range(num_joints):
                    joint_info = p.getJointInfo(self._robot, i)
                    joint_name = joint_info[1].decode("UTF-8")
                    link_name = joint_info[12].decode("UTF-8")
                    if joint_name == name:
                        self._robot_joints.append(i)
                    self._link_names.append(link_name)
                    if joint_name in self._actuated_wheels:
                        self._wheel_joints.append(i)
                    if joint_name in self._steering_links:
                        self._steering_joints.append(i)

    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:
        """ Reset simulation and add robot """
        logging.warning(
            "The argument 'mount_position' and 'mount_orientation' are \
ignored for bicycle models."
        )
        if hasattr(self, "_robot"):
            p.removeBody(self._robot)
        base_orientation = p.getQuaternionFromEuler([0, 0, pos[2]])
        spawn_position = self._spawn_offset
        spawn_position[0:2] += pos[0:2]
        self._robot = p.loadURDF(
            fileName=self._urdf_file,
            basePosition=spawn_position,
            baseOrientation=base_orientation,
            flags=p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT,
            globalScaling=self._scaling,
        )
        self.set_joint_names()
        self.extract_joint_ids()
        self.read_limits()
        # set base velocity
        self.update_state()
        self._integrated_forward_velocity = vel[0]

    def read_limits(self) -> None:
        self._limit_pos_j = np.zeros((2, self.ns()))
        self._limit_vel_j = np.zeros((2, self.ns()))
        self._limit_tor_j = np.zeros((2, self.n()))
        self._limit_acc_j = np.zeros((2, self.n()))
        self._limit_pos_steering = np.zeros(2)
        if len(self._steering_joints) > 0:
            joint = self._urdf_robot.robot.joints[self._steering_joints[1] - 1]
            if joint.limit is not None:
                self._limit_pos_steering[0] = joint.limit.lower - 0.1
                self._limit_pos_steering[1] = joint.limit.upper + 0.1
        self._limit_vel_forward_j = np.array([[-40., -10.], [40., 10.]])
        self._limit_pos_j[0, 0:3] = np.array([-1000., -1000., -2 * np.pi])
        self._limit_pos_j[1, 0:3] = np.array([1000., 1000, 2 * np.pi])
        self._limit_vel_j[0, 0:3] = np.array([-40., -40., -10.])
        self._limit_vel_j[1, 0:3] = np.array([40., 40., 10.])
        self.set_acceleration_limits()

    def set_joint_names(self):
        self._joint_names = self._steering_links + self._actuated_wheels

    def set_acceleration_limits(self):
        acc_limit = np.array([1.0, 1.0])
        self._limit_acc_j[0, :] = -acc_limit
        self._limit_acc_j[1, :] = acc_limit

    def check_state_new(self, pos: np.ndarray, vel: np.ndarray):
        if (
            not isinstance(pos, np.ndarray)
            or not pos.size == self.n() + 1
        ):
            pos = np.zeros(self.n() + 1)
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel

    def check_state(self, pos: np.ndarray, vel: np.ndarray) -> tuple:
        """Filters state of the robot and returns a valid state."""

        if not isinstance(pos, np.ndarray) or not pos.size == self.ns():
            pos = np.zeros(self.ns())
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel

    def get_observation_space(self) -> gym.spaces.Dict:
        """Gets the observation space for a bicycle model.

        The observation space is represented as a dictonary. `x` and `xdot`
        denote the configuration position and velocity, `vel` for forward and
        angular velocity and `steering` is the current steering position.
        """
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict(
                    {
                        "position": gym.spaces.Box(
                            low=self._limit_pos_j[0, :],
                            high=self._limit_pos_j[1, :],
                            dtype=float,
                        ),
                        "steering": gym.spaces.Box(
                            low=self._limit_pos_steering[0],
                            high=self._limit_pos_steering[1],
                            shape=(1,),
                            dtype=float,
                        ),
                        "velocity": gym.spaces.Box(
                            low=self._limit_vel_j[0, :],
                            high=self._limit_vel_j[1, :],
                            dtype=float,
                        ),
                        "forward_velocity": gym.spaces.Box(
                            low=self._limit_vel_forward_j[0, :],
                            high=self._limit_vel_forward_j[1, :],
                            dtype=float,
                        ),
                    }
                )
            }
        )

    def get_velocity_spaces(self) -> tuple:
        """Get observation space and action space when using velocity
        control."""
        ospace = self.get_observation_space()
        uu = self._limit_vel_forward_j[1, :]
        ul = self._limit_vel_forward_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def apply_velocity_action(self, vels: np.ndarray) -> None:
        """Applies velocities to forward motion and sets steering angle."""
        for steering_joint in self._steering_joints:
            p.setJointMotorControl2(
                self._robot,
                steering_joint,
                controlMode=p.POSITION_CONTROL,
                targetPosition=vels[1],
            )
        for joint in self._wheel_joints:
            p.setJointMotorControl2(
                self._robot,
                joint,
                controlMode=p.VELOCITY_CONTROL,
                targetVelocity=vels[0] / (self._wheel_radius * self._scaling),
            )

    def apply_acceleration_action(self, accs: np.ndarray, dt: float) -> None:
        """Applies acceleration action to the robot.

        The acceleration action relies on integration of the velocity signal
        and applies velocities at the end. The integrated velocities are
        clipped to avoid very large velocities.
        """
        self._integrated_forward_velocity += dt * accs[0]
        actions = np.array([self._integrated_forward_velocity, accs[1]])
        self.apply_velocity_action(actions)

    def apply_torque_action(self, torques: np.ndarray) -> None:
        raise NotImplementedError("Torque action is not available for prius.")

    def correct_base_orientation(self, pos_base: np.ndarray) -> np.ndarray:
        """Corrects base orientation to be within the interval (-pi , pi].
        """
        if self._facing_direction == '-y':
            pos_base[2] -= np.pi/2
        elif self._facing_direction == 'y':
            pos_base[2] += np.pi/2
        elif self._facing_direction == '-x':
            pos_base[2] += np.pi
        if pos_base[2] < -np.pi:
            pos_base[2] += 2 * np.pi
        if pos_base[2] > np.pi:
            pos_base[2] -= 2 * np.pi
        return pos_base

    def update_state(self) -> None:
        """Updates the robot state.

        The robot state is stored in the dictonary self.state.  There, the key
        x refers to the translational and rotational position in the world
        frame. The key xdot refers to the translation and rotational velocity
        in the world frame. The key vel is the current forward and angular
        velocity. For a bicycle model we additionally store information about
        the sterring behind the key steering.
        """
        # base position
        link_state = p.getLinkState(self._robot, 0, computeLinkVelocity=1)
        pos_base = np.array(
            [
                link_state[0][0],
                link_state[0][1],
                p.getEulerFromQuaternion(link_state[1])[2],
            ]
        )
        self.correct_base_orientation(pos_base)
        vel_base = np.array(
            [link_state[6][0], link_state[6][1], link_state[7][2]]
        )
        # wheel velocities
        vel_wheels = p.getJointStates(self._robot, self._wheel_joints[2:4])
        v_right = vel_wheels[0][1]
        v_left = vel_wheels[1][1]
        vel = np.array(
            [
                0.5 * (v_right + v_left) * self._scaling * self._wheel_radius,
                vel_base[2],
            ]
        )
        pos, _, _, _ = p.getJointState(self._robot, self._steering_joints[1])
        steering_pos = np.array([pos])
        self.state = {
            "joint_state": {
                "position": pos_base,
                "forward_velocity": vel,
                "velocity": vel_base,
                "steering": steering_pos,
            }
        }



import numpy as np
import time
import logging
from typing import List, Optional, Union
import xml.etree.ElementTree as ET

import gymnasium as gym
from gymnasium import Env, utils
import mujoco
from dm_control import mjcf
from urdfenvs.sensors.lidar import Lidar
from urdfenvs.sensors.sensor import Sensor

from urdfenvs.urdf_common.urdf_env import (
    check_observation,
    WrongObservationError,
)

from urdfenvs.generic_mujoco.generic_mujoco_robot import GenericMujocoRobot
from gymnasium.envs.mujoco.mujoco_rendering import MujocoRenderer
from mpscenes.obstacles.collision_obstacle import CollisionObstacle
from mpscenes.goals.sub_goal import SubGoal

DEFAULT_CAMERA_CONFIG = {
    "trackbodyid": -1,
    "distance": 4.0,
}

DEFAULT_SIZE = 480


class LinkIdNotFoundError(Exception):
    pass


class GenericMujocoEnv(Env):
    metadata = {
        "render_modes": [
            "human",
            "rgb_array",
            "depth_array",
        ],
        "render_fps": 20,
    }
    _t: float
    _number_movable_obstacles: int

    def __init__(
        self,
        robots: List[GenericMujocoRobot],
        obstacles: List[CollisionObstacle],
        goals: List[SubGoal],
        sensors: Optional[List[Sensor]] = None,
        render: Optional[Union[str, bool]] = None,
        frame_skip: int = 5,
        width: int = DEFAULT_SIZE,
        height: int = DEFAULT_SIZE,
        camera_id: Optional[int] = None,
        camera_name: Optional[str] = None,
        enforce_real_time: Optional[bool] = None,
    ) -> None:

        utils.EzPickle.__init__(self)
        self._xml_file = robots[0].xml_file
        self._done: bool = False
        self._obstacles = obstacles
        self._goals = goals
        if not sensors:
            self._sensors = []
        else:
            self._sensors = sensors
        self._number_movable_obstacles = 0
        self.add_scene()
        self._t = 0.0

        render_mode = render
        if render is True:
            render_mode = "human"
        if render is False:
            render_mode = None
        if enforce_real_time and render_mode == "human":
            self._enforce_real_time = True
        else:
            self._enforce_real_time = False

        self.width = width
        self.height = height

        self.frame_skip = frame_skip
        self._initialize_simulation()

        for sensor in self._sensors:
            sensor.set_data(self.data)

        assert self.metadata["render_modes"] == [
            "human",
            "rgb_array",
            "depth_array",
        ], self.metadata["render_modes"]
        if "render_fps" in self.metadata:
            assert (
                int(np.round(1.0 / self.dt)) == self.metadata["render_fps"]
            ), f'Expected value: {int(np.round(1.0 / self.dt))}, Actual value: {self.metadata["render_fps"]}'

        self.observation_space = self.get_observation_space()
        self.action_space = self.get_action_space()

        self.render_mode = render_mode
        self.camera_name = camera_name
        self.camera_id = camera_id

        self.mujoco_renderer = MujocoRenderer(
            self.model,
            self.data,
            DEFAULT_CAMERA_CONFIG,
            height=DEFAULT_SIZE,
            width=DEFAULT_SIZE,
        )

    def render(self):
        return self.mujoco_renderer.render(self.render_mode)

    def get_observation_space(self) -> gym.spaces.Dict:
        """Get observation space."""
        observation_space_as_dict = dict(
            gym.spaces.Dict(
                {
                    "joint_state": gym.spaces.Dict(
                        {
                            "position": gym.spaces.Box(
                                low=self.joint_limits()[:, 0],
                                high=self.joint_limits()[:, 1],
                                dtype=float,
                            ),
                            "velocity": gym.spaces.Box(
                                low=np.ones_like(self.joint_limits()[:, 0])
                                * -10.0,
                                high=np.ones_like(self.joint_limits()[:, 0])
                                * 10.0,
                                dtype=float,
                            ),
                        }
                    )
                }
            )
        )
        for sensor in self._sensors:
            observation_space_as_dict.update(
                sensor.get_observation_space(
                    self.obstacles_dict, self.goals_dict
                )
            )
        observation_space = gym.spaces.Dict(
            {"robot_0": gym.spaces.Dict(observation_space_as_dict)}
        )
        return observation_space

    def get_action_space(self) -> np.ndarray:
        return gym.spaces.Box(
            low=self.actuator_limits()[:, 0],
            high=self.actuator_limits()[:, 1],
            dtype=float,
        )

    def joint_limits(self) -> np.ndarray:
        return self.model.jnt_range[: self.nq]

    def actuator_limits(self) -> np.ndarray:
        return self.model.actuator_ctrlrange

    def velocity_limits(self) -> np.ndarray:
        return self.model.actuator_ctrlrange

    def _initialize_simulation(
        self,
    ):

        file_name = self._xml_file.split("/")[-1]
        mjcf.export_with_assets(
            self._model_dm,
            "xml_model",
            out_file_name=file_name,
        )
        self.model = mujoco.MjModel.from_xml_path(f"xml_model/{file_name}")

        self.model.body_pos[0] = [0, 1, 1]
        self.model.vis.global_.offwidth = self.width
        self.model.vis.global_.offheight = self.height
        self.data = mujoco.MjData(self.model)

    @property
    def dt(self) -> float:
        return self.model.opt.timestep * self.frame_skip

    @property
    def t(self) -> float:
        return self._t

    @property
    def obstacles_dict(self) -> dict:
        return {k: v for k, v in enumerate(self._obstacles)}

    @property
    def goals_dict(self) -> dict:
        return {k: v for k, v in enumerate(self._goals)}

    def update_obstacles_position(self):

        non_movable_obstacles = [
            obstacle for obstacle in self._obstacles if not obstacle.movable()
        ]
        for i, obstacle in enumerate(non_movable_obstacles):
            self.data.mocap_pos[i] = obstacle.position(t=self.t)

    def update_goals_position(self):
        for i, goal in enumerate(self._goals):
            self.data.site_xpos[i] = goal.position(t=self.t)

    def step(self, action: np.ndarray):
        step_start = time.perf_counter()
        self._t += self.dt
        truncated = False
        info = {}
        reward = 0
        if not self.action_space.contains(action):
            self._done = True
            info = {"action_limits": f"{action} not in {self.action_space}"}

        self.do_simulation(action, self.frame_skip)
        for contact in self.data.contact:
            body1 = self.model.geom(contact.geom1).name
            body2 = self.model.geom(contact.geom2).name
            if "movable" in body1 or "movable" in body2:
                continue

            message = (
                f"Collision occured at {round(self.t, 2)} "
                f"between {body1} and obstacle "
                f"with id {body2}"
            )
            info = {"Collision": message}
            self._done = True
        self.update_obstacles_position()
        self.update_goals_position()
        if self.render_mode == "human":
            self.render()

        ob = self._get_obs()
        if not self.observation_space.contains(ob):
            try:
                check_observation(self.observation_space, ob)
            except WrongObservationError as e:
                self._done = True
                info = {"observation_limits": str(e)}
        step_end = time.perf_counter()
        step_time = step_end - step_start
        if self._enforce_real_time:
            sleep_time = max(0.0, self.dt - step_time)
            time.sleep(sleep_time)
        step_final_end = time.perf_counter()
        total_step_time = step_final_end - step_start
        real_time_factor = self.dt / total_step_time
        logging.info(f"Real time factor {real_time_factor}")
        return (
            ob,
            reward,
            self._done,
            truncated,
            info,
        )

    def reset(
        self,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
        pos: Optional[np.ndarray] = None,
        vel: Optional[np.ndarray] = None,
    ):
        super().reset(seed=seed, options=options)
        if options and options.get("randomize_obstacles", False):
            for obstacle in self._obstacles:
                obstacle.shuffle()
        if options and options.get("randomize_goals", False):
            for goal in self._goals:
                goal.shuffle()
        if pos is not None:
            qpos = pos
        else:
            qpos = (self.joint_limits()[:, 1] + self.joint_limits()[:, 0]) / 2
        if vel is not None:
            qvel = vel
        else:
            qvel = np.zeros(self.nv)
        self.set_state(qpos, qvel)
        self._t = 0.0
        return self._get_obs(), {}

    def set_state(self, qpos, qvel):
        assert qpos.shape == (self.nq,) and qvel.shape == (self.nv,)
        self.data.qpos[: self.nq] = np.copy(qpos)
        self.data.qvel[: self.nv] = np.copy(qvel)
        if self.model.na == 0:
            self.data.act[:] = None
        mujoco.mj_forward(self.model, self.data)

    @property
    def nu(self) -> int:
        return self.model.nu

    @property
    def nq(self) -> int:
        return self.model.nq - 7 * self._number_movable_obstacles

    @property
    def nv(self) -> int:
        return self.model.nv - 6 * self._number_movable_obstacles

    def do_simulation(self, ctrl, n_frames) -> None:
        """
        Step the simulation n number of frames and applying a control action.
        """
        # Check control input is contained in the action space
        if np.array(ctrl).shape != (self.model.nu,):
            raise ValueError(
                f"Action dimension mismatch. Expected {(self.model.nu,)}, found {np.array(ctrl).shape}"
            )
        self._step_mujoco_simulation(ctrl, n_frames)

    def _step_mujoco_simulation(self, ctrl, n_frames):
        self.data.ctrl[:] = ctrl

        mujoco.mj_step(self.model, self.data, nstep=n_frames)

        # As of MuJoCo 2.0, force-related quantities like cacc are not computed
        # unless there's a force sensor in the model.
        # See https://github.com/openai/gym/issues/1541
        mujoco.mj_rnePostConstraint(self.model, self.data)

    def add_scene(self) -> None:
        self._model_dm = mjcf.from_path(self._xml_file)
        for obstacle in self._obstacles:
            self.add_obstacle(obstacle)
        for sub_goal in self._goals:
            self.add_sub_goal(sub_goal)
        for sensor in self._sensors:
            if isinstance(sensor, Lidar):
                self.add_range_finder(sensor)

    def add_range_finder(self, sensor: Lidar) -> None:
        try:
            lidar_body = self._model_dm.find("body", sensor._link_name)
        except Exception as e:
            print(e)
            raise LinkIdNotFoundError(
                f"Link '{sensor._link_name}' not found in xml. It might be in an include."
            )

        for i in range(sensor._nb_rays):
            angle = i / sensor._nb_rays * np.pi * 2
            start_position = np.array(
                [
                    np.cos(angle) * 0.00,
                    np.sin(angle) * 0.00,
                    0.0,
                ]
            )
            end_position = np.array(
                [
                    np.cos(angle) * 0.01,
                    np.sin(angle) * 0.01,
                    0.0,
                ]
            )
            fromto_string = " ".join(map(str, start_position))
            fromto_string += " " + " ".join(map(str, end_position))

            site_values = {
                "name": f"{sensor.name()}_rf_{i}",
                "type": "capsule",
                "size": "0.01",
                "fromto": fromto_string,
            }
            lidar_body.add("site", **site_values)

            rangefinder_values = {
                "name": f"{sensor.name()}_{i}",
                "site": f"{sensor.name()}_rf_{i}",
                "cutoff": str(sensor._ray_length + 0.01 / 2),
            }
            self._model_dm.sensor.add("rangefinder", **rangefinder_values)

    def add_obstacle(self, obst: CollisionObstacle) -> None:
        if obst.type() == "sphere":
            size = str(obst.size()[0])
        else:
            size = " ".join([str(i / 2) for i in obst.size()])

        geom_values = {
            "name": obst.name(),
            "type": obst.type(),
            "rgba": " ".join([str(i) for i in obst.rgba()]),
            "size": size,
        }
        obstacle_body = self._model_dm.worldbody.add(
            "body",
            name=obst.name(),
            pos=obst.position().tolist(),
            mocap=not obst.movable(),
        )
        if obst.movable():
            obstacle_body.add("freejoint", name=f"freejoint_{obst.name()}")
            self._number_movable_obstacles += 1
        obstacle_body.add("geom", **geom_values)

    def add_sub_goal(self, sub_goal: SubGoal) -> None:
        position = np.zeros(3)
        for index in sub_goal.indices():
            position[index] = sub_goal.position()[index]
        geom_values = {
            "name": sub_goal.name(),
            "type": "sphere",
            "rgba": "0 1 0 0.3",
            "pos": " ".join([str(i) for i in position.tolist()]),
            "size": str(sub_goal.epsilon()),
        }
        self._model_dm.worldbody.add("site", **geom_values)

    def _get_obs(self):
        observation = {
            "joint_state": {
                "position": self.data.qpos[0 : self.nq],
                "velocity": self.data.qvel[0 : self.nv],
            }
        }
        for sensor in self._sensors:
            observation.update(
                {
                    sensor.name(): sensor.sense(
                        0, self.obstacles_dict, self.goals_dict, self.t
                    )
                }
            )
        return {"robot_0": observation}

    def close(self):
        if self.mujoco_renderer is not None:
            self.mujoco_renderer.close()

    def xml(self) -> str:
        """Return the xml string of the model.

        As the model from the xml file may be changed during the simulation,
        by adding obstacles, goals, sensors, etc., this method ensures that
        the updated model can be saved as an xml file.

        """
        return self._model_dm.to_xml_string()



"""Abstract class for sensor."""
from typing import Optional
from abc import abstractmethod

import numpy as np

from urdfenvs.sensors.physics_engine_interface import PhysicsEngineInterface, PybulletInterface, MujocoInterface

class Sensor():
    """Abstract sensor class.

    This class serves as a blueprint for sensors. Every sensor must
    inherit from this class and implement the abstract methods.
    """

    _name: str
    _variance: float
    _plotting_interval: int
    _physics_engine: PhysicsEngineInterface

    def __init__(
        self,
        name: str,
        variance: float = 0.0,
        plotting_interval: int = -1,
        physics_engine_name: str = 'pybullet',
    ):
        self._name = name
        self._variance = variance
        self._plotting_interval = plotting_interval
        self._call_counter = plotting_interval - 1
        if physics_engine_name == 'pybullet':
            self._physics_engine = PybulletInterface()
        elif physics_engine_name == 'mujoco':
            self._physics_engine = MujocoInterface()

    def name(self):
        return self._name

    def set_data(self, data):
        self._physics_engine.set_data(data)

    def add_noise(self, exact_data: np.ndarray, limits: Optional[np.ndarray] = None):
        """Add noise to the exact data."""
        noisy_data = np.random.normal(exact_data, self._variance)
        used_limits = limits if limits is not None else self._observation_limits
        
        if np.all(exact_data >= used_limits[0]) and np.all(exact_data <= used_limits[1]):
            clipped = np.clip(noisy_data, used_limits[0], used_limits[1])
            return clipped
        else:
            return noisy_data


    @abstractmethod
    def get_observation_size(self):
        pass

    @abstractmethod
    def get_observation_space(self, obstacles: dict, goals: dict):
        pass

    @abstractmethod
    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        pass



from gymnasium import register
from urdfenvs.generic_mujoco.generic_mujoco_env import GenericMujocoEnv
register(
    id='generic-mujoco-env-v0',
    entry_point='urdfenvs.generic_mujoco.generic_mujoco_env:GenericMujocoEnv'
)



import os

class GenericMujocoRobot:
    """GenericMujocoRobot."""

    _xml_file: str

    def __init__(self, xml_file: str, mode: str = "vel"):
        """Constructor for generic robot.

        Parameters
        ----------

        xml_file: str :
            Name of xml file.
        mode: str:
            Control mode. Note that the mode is not used in mujoco as it is
            implicitely defined by the actuators.
        """
        if not os.path.exists(xml_file):
            raise FileNotFoundError(f"Requested xml {xml_file} not found")
        else:
            self._xml_file = xml_file

    @property
    def xml_file(self) -> str:
        return self._xml_file
